@@ -7,6 +7,7 @@ from .cssselectpatch import JQueryTranslator
 from copy import deepcopy
 from lxml import etree
 import lxml.html
+import inspect
 import sys
 
 PY3k = sys.version_info >= (3,)
@@ -156,7 +157,9 @@ class PyQuery(list):
         if kwargs:
             # specific case to get the dom
             if 'filename' in kwargs:
-                html = open(kwargs['filename'])
+                fd = open(kwargs['filename'])
+                html = fd.read()
+                fd.close()
             elif 'url' in kwargs:
                 url = kwargs.pop('url')
                 if 'opener' in kwargs:
@@ -207,7 +210,6 @@ class PyQuery(list):
                     elements = fromstring(context, self.parser)
                 except Exception:
                     raise
-                    raise ValueError(repr(context))
             elif isinstance(context, self.__class__):
                 # copy
                 elements = context[:]
@@ -257,9 +259,23 @@ class PyQuery(list):
         return self.__class__(self[:] + other[:])
 
     def extend(self, other):
+        """Extend with anoter PyQuery object"""
         assert isinstance(other, self.__class__)
         self._extend(other[:])
 
+    def items(self, selector=None):
+        """Iter over elements. Return PyQuery objects:
+
+            >>> d = PyQuery('<div><span>foo</span><span>bar</span></div>')
+            >>> [i.text() for i in d.items('span')]
+            ['foo', 'bar']
+            >>> [i.text() for i in d('span').items()]
+            ['foo', 'bar']
+        """
+        elems = selector and self(selector) or self
+        for elem in elems:
+            yield self.__class__(elem)
+
     def __str__(self):
         """xml representation of current nodes::
 
@@ -475,6 +491,19 @@ class PyQuery(list):
                 result.append(current)
         return self.__class__(result, **dict(parent=self))
 
+    def contents(self):
+        """
+        Return contents (with text nodes):
+
+            >>> d = PyQuery('hello <b>bold</b>')
+            >>> d.contents()  # doctest: +ELLIPSIS
+            ['hello ', <Element b at ...>]
+        """
+        results = []
+        for elem in self:
+            results.extend(elem.xpath('child::text()|child::*'))
+        return self.__class__(results, **dict(parent=self))
+
     def filter(self, selector):
         """Filter elements in self using selector (string or function):
 
@@ -487,15 +516,19 @@ class PyQuery(list):
             [<p>]
             >>> d('p').filter(lambda i: PyQuery(this).text() == 'Hi')
             [<p.hello>]
+            >>> d('p').filter(lambda i, this: PyQuery(this).text() == 'Hi')
+            [<p.hello>]
         """
         if not hasattr(selector, '__call__'):
             return self._filter_only(selector, self)
         else:
             elements = []
+            args = inspect.getargspec(callback).args
             try:
                 for i, this in enumerate(self):
-                    func_globals(selector)['this'] = this
-                    if callback(selector, i):
+                    if len(args) == 1:
+                        func_globals(selector)['this'] = this
+                    if callback(selector, i, this):
                         elements.append(this)
             finally:
                 f_globals = func_globals(selector)
@@ -718,9 +751,8 @@ class PyQuery(list):
         """
         for tag in self:
             values = value.split(' ')
-            classes = set((tag.get('class') or '').split())
-            classes = classes.union(values)
-            classes.difference_update([''])
+            classes = (tag.get('class') or '').split()
+            classes += [v for v in values if v not in classes]
             tag.set('class', ' '.join(classes))
         return self
 
@@ -750,12 +782,12 @@ class PyQuery(list):
 
         """
         for tag in self:
-            values = set(value.split(' '))
-            classes = set((tag.get('class') or '').split())
-            values_to_add = values.difference(classes)
-            classes.difference_update(values)
-            classes = classes.union(values_to_add)
-            classes.difference_update([''])
+            values = value.split(' ')
+            classes = (tag.get('class') or '').split()
+            values_to_add = [v for v in values if v not in classes]
+            values_to_del = [v for v in values if v in classes]
+            classes = [v for v in classes if v not in values_to_del]
+            classes += values_to_add
             tag.set('class', ' '.join(classes))
         return self
 
@@ -836,7 +868,7 @@ class PyQuery(list):
             'Youhou'
 
         """
-        return self.attr('value', value)
+        return self.attr('value', value) or None
 
     def html(self, value=no_default, **kwargs):
         """Get or set the html representation of sub nodes.
@@ -947,7 +979,7 @@ class PyQuery(list):
             text = []
 
             def add_text(tag, no_tail=False):
-                if tag.text:
+                if tag.text and not isinstance(tag, lxml.etree._Comment):
                     text.append(tag.text)
                 for child in tag.getchildren():
                     add_text(child)
@@ -1193,8 +1225,7 @@ class PyQuery(list):
     def clone(self):
         """return a copy of nodes
         """
-        self[:] = [deepcopy(tag) for tag in self]
-        return self
+        return PyQuery([deepcopy(tag) for tag in self])
 
     def empty(self):
         """remove nodes content
